/* Copyright 2017-2025
 * Kaz Kylheku <kaz@kylheku.com>
 * Vancouver, Canada
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <wchar.h>
#include <limits.h>
#include <float.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <stdio.h>
#include <wchar.h>
#include "config.h"
#if HAVE_ZLIB
#include <zlib.h>
#endif
#include "lib.h"
#include "gc.h"
#include "itypes.h"
#include "signal.h"
#include "unwind.h"
#include "eval.h"
#include "utf8.h"
#define STREAM_IMPL
#include "stream.h"
#include "arith.h"
#include "txr.h"
#include "buf.h"

#define BUF_BORROWED UINT_PTR_MAX

#define min(a, b) ((a) < (b) ? (a) : (b))

static struct buf *buf_handle(val buf, val ctx);

static ucnum buf_check_len(val len, val self)
{
  if (minusp(len)) {
    uw_throwf(error_s, lit("~a: negative length ~s specified"),
              self, len, nao);
  } else {
    ucnum l = c_unum(len, self);

    if (l == BUF_BORROWED)
      uw_throwf(error_s, lit("~a: reserved length value ~s specified"),
                self, len, nao);

    return l;
  }
}

static ucnum buf_check_alloc_size(val alloc_size, ucnum len, val self)
{
  ucnum as = c_unum(alloc_size, self);

  if (as == BUF_BORROWED)
      uw_throwf(error_s, lit("~a: reserved alloc size value ~s specified"),
                self, alloc_size, nao);

  if (as < len)
    uw_throwf(error_s, lit("~a: alloc size size ~s lower than length"),
              self, alloc_size, nao);

  return as;
}

static ucnum buf_check_index(struct buf *b, val index_in, val self)
{
  val index = minusp(index_in) ? plus(unum(b->len), index_in) : index_in;
  ucnum ix;

  if (minusp(index)) {
    uw_throwf(error_s, lit("~a: negative byte index ~s specified"),
              self, index, nao);
  }

  ix = c_unum(index, self);

  if (ix >= BUF_BORROWED - 1) {
    uw_throwf(error_s, lit("~a: index ~s disallowed"),
              self, index, nao);
  }

  return ix;
}

NORETURN static void err_oflow(val self)
{
  uw_throwf(error_s, lit("~a: array size overflow"), self, nao);
}

static void prepare_pattern(mem_t **ppat, mem_t **pfree, mem_t pbuf[SIZEOF_PTR],
                            ucnum *plen, int *left_p, val pat, val self)
{
  *left_p = 0;
  *pfree = 0;

  switch (type(pat)) {
  case CHR:
  case NUM:
    {
      cnum pn = c_num(pat, self);
      unsigned char *fill = pbuf + SIZEOF_PTR;

      if (pn < 0) {
        *left_p = 1;
        pn = -pn;
      }

      do {
        *--fill = pn & 0xff;
        pn >>= 8;
      } while (fill > pbuf && pn);

      *plen = pbuf + SIZEOF_PTR - fill;
      *ppat = fill;
    }
    break;
  case BGNUM:
    {
      if (minusp(pat)) {
        *left_p = 1;
        pat = neg(pat);
      }

      {
        mp_int *m = mp(pat);
        ucnum pl = mp_unsigned_bin_size(m);
        mem_t *p = chk_malloc(pl);
        mp_to_unsigned_bin(m, p);
        *ppat = *pfree = p;
        *plen = pl;
      }
    }
    break;
  case BUF:
    {
      struct buf *pb = buf_handle(pat, self);
      ucnum l = pb->len;

      if (l > 0) {
        *ppat = pb->data;
        *plen = l;
      } else {
        *ppat = pbuf;
        *plen = 1;
      }
    }
    break;
  default:
    unsup_obj(self, pat);
  }
}

val make_buf(val len, val init_pat, val alloc_size)
{
  val self = lit("make-buf");
  val pat = default_arg(init_pat, zero);
  ucnum blen = buf_check_len(len, self);
  val alloc = if3(null_or_missing_p(alloc_size), len, alloc_size);
  ucnum size = buf_check_alloc_size(alloc, blen, self);
  mem_t *data = if3(pat == zero && size == blen,
                    chk_calloc(size, 1),
                    chk_malloc(size));
  val obj = make_obj();
  cnum iv;

  obj->b.type = BUF;
  obj->b.data = data;
  obj->b.len = blen;
  obj->b.size = size;

  if ((is_num(pat) || is_chr(pat)) &&
      -256 < (iv = c_num(pat, self)) && iv < 256)
  {
    if (iv != 0 || size != blen)
      memset(data, convert(unsigned char, iv), blen);
  } else {
    mem_t *p = 0, *fr = 0;
    unsigned char pbuf[SIZEOF_PTR] = { 0 };
    int left = 0;
    ucnum pl, l = size, i, j;

    prepare_pattern(&p, &fr, pbuf, &pl, &left, pat, self);

    if (left) {
      for (i = 0, j = 0; i < l; i++) {
        data[i] = p[j++];
        if (j >= pl)
          j = 0;
      }
    } else {
      for (i = l - 1, j = pl; i != convert(ucnum, -1); --i) {
        data[i] = p[--j];
        if (j == 0)
          j = pl;
      }
    }

    free(fr);
    gc_hint(pat);
  }

  return obj;
}

val make_buf_fast(ucnum len, int init_val, ucnum alloc_size)
{
  val obj = make_obj();
  bug_unless (alloc_size != BUF_BORROWED);
  if (alloc_size == 0)
    alloc_size = len;
  obj->b.type = BUF;
  obj->b.data = chk_malloc(alloc_size);
  obj->b.len = len;
  obj->b.size = alloc_size;
  memset(obj->b.data, init_val, len);
  return obj;
}

val bufp(val object)
{
  return tnil(type(object) == BUF);
}

val init_borrowed_buf(obj_t *obj, ucnum len, mem_t *data)
{
  obj->b.type = BUF;
  obj->b.data = data;
  obj->b.len = len;
  obj->b.size = BUF_BORROWED;

  return obj;
}

val make_borrowed_buf(ucnum len, mem_t *data)
{
  return init_borrowed_buf(make_obj(), len, data);
}

val make_duplicate_buf(ucnum len, mem_t *data, val self)
{
  if (len < BUF_BORROWED) {
    val obj = make_obj();

    obj->b.type = BUF;
    obj->b.data = chk_copy_obj(data, len);
    obj->b.len = len;
    obj->b.size = len;

    return obj;
  } else {
    err_oflow(self);
  }
}

val make_owned_buf(ucnum len, mem_t *data, val self)
{
  if (len < BUF_BORROWED) {
    val buf = make_borrowed_buf(len, data);
    buf->b.size = len;
    return buf;
  } else {
    err_oflow(self);
  }
}

static val make_ubuf(ucnum len)
{
  mem_t *data = chk_malloc(len);
  val buf = make_borrowed_buf(len, data);
  buf->b.size = len;
  return buf;
}

static struct buf *buf_handle(val buf, val ctx)
{
  if (type(buf) == BUF)
    return &buf->b;
  uw_throwf(error_s, lit("~a: ~s isn't a buffer"),
            ctx, buf, nao);
}

INLINE struct buf *us_buf_handle(val buf)
{
  return &buf->b;
}

val copy_buf(val buf)
{
  val self = lit("copy-buf");
  struct buf *b = buf_handle(buf, self);
  return if3(b->size != BUF_BORROWED,
             make_duplicate_buf(b->len, b->data, self),
             make_borrowed_buf(b->len, b->data));
}

static void buf_shrink(struct buf *b)
{
  ucnum len = b->len;

  if (len == 0)
    len = 1; /* avoid zero size realloc */

  if (len != b->size) {
    b->data = chk_realloc(b->data, len);
    b->size = b->len;
  }
}

val buf_trim(val buf)
{
  val self = lit("buf-trim");
  struct buf *b = buf_handle(buf, self);
  ucnum oldsize = b->size;
  if (oldsize == BUF_BORROWED)
    uw_throwf(error_s, lit("~a: ~s is a fixed buffer"),
              self, buf, nao);
  buf_shrink(b);
  return unum(oldsize);
}

static val buf_do_set_len(val buf, struct buf *b, ucnum len,
                          val init_pat, val self)
{
  ucnum olen = b->len;
  ucnum oldsize = b->size, size = oldsize;

  if (b->size == BUF_BORROWED)
    uw_throwf(error_s, lit("~a: ~s is a fixed buffer"),
              self, buf, nao);

  b->len = len;

  if (size < len) {
    if (size > INT_PTR_MAX - INT_PTR_MAX / 5) {
      size = INT_PTR_MAX;
    } else {
      size = size + size / 4;
      if (size < len)
        size = len;
    }
  }

  if (size > oldsize) {
    b->data = chk_realloc(b->data, size);
    b->size = size;
  }

  if (len > olen) {
    val pat = default_arg(init_pat, zero);
    cnum iv;
    if ((is_num(pat) || is_chr(pat)) &&
        -256 < (iv = c_num(pat, self)) && iv < 256)
    {
      memset(b->data + olen, convert(unsigned char, iv), len - olen);
    } else {
      mem_t *p = 0, *fr = 0;
      unsigned char pbuf[SIZEOF_PTR] = { 0 };
      int left = 0;
      ucnum pl, from = olen, l = len, i, j;

      prepare_pattern(&p, &fr, pbuf, &pl, &left, pat, self);

      if (left) {
        for (i = from, j = from % pl; i < l; i++) {
          b->data[i] = p[j++];
          if (j >= pl)
            j = 0;
        }
      } else {
        for (i = l - 1, j = pl; i != from - 1; --i) {
          b->data[i] = p[--j];
          if (j == 0)
            j = pl;
        }
      }

      free(fr);
      gc_hint(pat);
    }
  }

  return unum(olen);
}

val buf_set_length(val buf, val len, val init_val)
{
  val self = lit("buf-set-len");
  struct buf *b = buf_handle(buf, self);
  ucnum l = buf_check_len(len, self);
  return buf_do_set_len(buf, b, l, init_val, self);
}

val buf_free(val buf)
{
  val self = lit("buf-free");
  struct buf *b = buf_handle(buf, self);
  if (b->size != BUF_BORROWED) {
    free(b->data);
    b->data = 0;
    return t;
  }
  return nil;
}

val length_buf(val buf)
{
  val self = lit("length-buf");
  struct buf *b = buf_handle(buf, self);
  return unum(b->len);
}

val buf_alloc_size(val buf)
{
  val self = lit("buf-alloc-size");
  struct buf *b = buf_handle(buf, self);
  return b->size == BUF_BORROWED ? nil : unum(b->size);
}

mem_t *buf_get(val buf, val self)
{
  struct buf *b = buf_handle(buf, self);
  return b->data;
}

val sub_buf(val buf, val from, val to)
{
  val self = lit("sub-buf");
  struct buf *b = buf_handle(buf, self);
  val len = unum(b->len);

  if (null_or_missing_p(from))
    from = zero;
  else if (from == t)
    from = len;
  else if (minusp(from)) {
    from = plus(from, len);
    if (to == zero)
      to = len;
  }

  if (null_or_missing_p(to) || to == t)
    to = len;
  else if (minusp(to))
    to = plus(to, len);

  from = max2(zero, min2(from, len));
  to = max2(zero, min2(to, len));

  if (ge(from, to)) {
    return make_buf_fast(0, 0, 0);
  } else if (from == zero && to == len) {
    return buf;
  } else {
    return make_duplicate_buf(buf_check_len(minus(to, from), self),
                              b->data + c_num(from, self), self);
  }
}

val replace_buf(val buf, val items, val from, val to)
{
  val self = lit("replace");
  val len = length_buf(buf);

  if (missingp(from)) {
    from = zero;
  } else if (from == t) {
    from = len;
  } else if (!integerp(from)) {
    seq_iter_t wh_iter, item_iter;
    ucnum offs = 0;
    ucnum l = c_unum(len, self), ol = l;
    val wh, item;
    seq_iter_init(self, &wh_iter, from);
    seq_iter_init(self, &item_iter, items);

    if (!missingp(to))
      uw_throwf(error_s,
                lit("~a: to-arg not applicable when from-arg is a list"),
                self, nao);

    while (seq_get(&item_iter, &item) && seq_get(&wh_iter, &wh)) {
      if (ge(wh, len))
        break;
      buf_put_uchar(buf, wh, item);
    }

    if (!opt_compat || opt_compat > 289) {
      while (seq_get(&wh_iter, &wh)) {
        cnum w = c_num(wh, self);

        if (w < 0)
          w += ol;

        if (w < 0)
          break;

        w -= offs;

        if (l <= INT_PTR_MAX && w >= convert(cnum, l))
          break;

        memmove(buf->b.data + w,
                buf->b.data + w + 1,
                l - w - 1);
        l--;
        offs++;

      }

      if (offs > 0)
        buf_set_length(buf, num_fast(l), zero);
    }

    return buf;
  } else if (minusp(from)) {
    from = plus(from, len);
    if (to == zero)
      to = len;
  }

  if (null_or_missing_p(to) || to == t)
    to = len;
  else if (minusp(to))
    to = plus(to, len);

  from = max2(zero, min2(from, len));
  to = max2(zero, min2(to, len));

  {
    val len_rep = minus(to, from);
    val len_it = length(items);

    if (gt(len_rep, len_it)) {
      val len_diff = minus(len_rep, len_it);
      ucnum t = c_unum(to, self);
      ucnum l = c_unum(len, self);

      memmove(buf->b.data + t - c_num(len_diff, self),
              buf->b.data + t,
              l - t);

      buf_set_length(buf, minus(len, len_diff), zero);
      to = plus(from, len_it);
    } else if (lt(len_rep, len_it)) {
      val len_diff = minus(len_it, len_rep);
      ucnum t = c_unum(to, self);
      ucnum l = c_unum(len, self);

      buf_set_length(buf, plus(len, len_diff), zero);

      memmove(buf->b.data + t + c_num(len_diff, self),
              buf->b.data + t,
              l - t);
      to = plus(from, len_it);
    }

    if (zerop(len_it))
      return buf;
    if (bufp(items)) {
      memmove(buf->b.data + c_num(from, self), items->b.data, c_num(len_it, self));
    } else {
      seq_iter_t item_iter;
      ucnum f = c_unum(from, self);
      ucnum t = c_unum(to, self);

      seq_iter_init(self, &item_iter, items);

      for (; f != t; f++) {
        val item = seq_geti(&item_iter);
        buf_put_uchar(buf, num(f), item);
      }
    }
  }

  return buf;
}

val buf_list(val list)
{
  val self = lit("buf-list");
  val buf = make_buf_fast(0, 0, 32);
  struct buf *b = us_buf_handle(buf);
  seq_iter_t iter;
  val elem;
  ucnum i;

  for (i = 0, seq_iter_init(self, &iter, list); seq_get(&iter, &elem); i++) {
    if (i + 1 < i)
      err_oflow(self);
    buf_do_set_len(buf, b, i + 1, elem, self);
  }

  return buf;
}

static void buf_move_bytes(val buf, val pos, mem_t *ptr, ucnum size, val self)
{
  struct buf *b = buf_handle(buf, self);
  ucnum p = buf_check_index(b, pos, self);
  ucnum req_len = p + size;
  if (req_len < p)
    err_oflow(self);
  if (req_len > b->len)
    buf_do_set_len(buf, b, req_len, nil, self);
  memmove(b->data + p, ptr, size);
}

val buf_put_buf(val dbuf, val pos, val sbuf)
{
  val self = lit("buf-put-buf");
  struct buf *sb = buf_handle(sbuf, self);
  buf_move_bytes(dbuf, pos, sb->data, sb->len, self);
  return sbuf;
}

static val compat_buf_put_buf(val dbuf, val pos, val sbuf)
{
  if (is_ptr(pos) && !integerp(pos))
    return buf_put_buf(dbuf, sbuf, pos);
  else
    return buf_put_buf(dbuf, pos, sbuf);
}

void buf_put_bytes(val buf, val pos, mem_t *ptr, ucnum size, val self)
{
  return buf_move_bytes(buf, pos, ptr, size, self);
}

#if HAVE_I8
val buf_put_i8(val buf, val pos, val num)
{
  val self = lit("buf-put-i8");
  struct buf *b = buf_handle(buf, self);
  ucnum p = buf_check_index(b, pos, self);
  i8_t v = c_i8(num, self);
  if (p >= b->len)
    buf_do_set_len(buf, b, p + 1, nil, self);
  b->data[p] = v;
  return num;
}

val buf_put_u8(val buf, val pos, val num)
{
  val self = lit("buf-put-u8");
  struct buf *b = buf_handle(buf, self);
  ucnum p = buf_check_index(b, pos, self);
  cnum v = c_u8(num, self);
  if (p >= b->len)
    buf_do_set_len(buf, b, p + 1, nil, self);
  b->data[p] = v;
  return num;
}
#endif

#if HAVE_I16
val buf_put_i16(val buf, val pos, val num)
{
  val self = lit("buf-put-i16");
  i16_t n = c_i16(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_u16(val buf, val pos, val num)
{
  val self = lit("buf-put-u16");
  u16_t n = c_u16(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}
#endif

#if HAVE_I32
val buf_put_i32(val buf, val pos, val num)
{
  val self = lit("buf-put-i32");
  i32_t n = c_i32(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_u32(val buf, val pos, val num)
{
  val self = lit("buf-put-u32");
  u32_t n = c_u32(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}
#endif

#if HAVE_I64
val buf_put_i64(val buf, val pos, val num)
{
  val self = lit("buf-put-i64");
  i64_t n = c_i64(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_u64(val buf, val pos, val num)
{
  val self = lit("buf-put-u64");
  u64_t n = c_u64(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}
#endif

val buf_put_char(val buf, val pos, val num)
{
  val self = lit("buf-put-char");
  struct buf *b = buf_handle(buf, self);
  ucnum p = buf_check_index(b, pos, self);
  char v = c_char(num, self);
  if (p >= b->len)
    buf_do_set_len(buf, b, p + 1, nil, self);
  b->data[p] = v;
  return num;
}

val buf_put_uchar(val buf, val pos, val num)
{
  val self = lit("buf-put-uchar");
  struct buf *b = buf_handle(buf, self);
  ucnum p = buf_check_index(b, pos, self);
  unsigned char v = c_uchar(num, self);
  if (p >= b->len)
    buf_do_set_len(buf, b, p + 1, nil, self);
  b->data[p] = v;
  return num;
}

val buf_put_short(val buf, val pos, val num)
{
  val self = lit("buf-put-short");
  short n = c_short(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_ushort(val buf, val pos, val num)
{
  val self = lit("buf-put-ushort");
  unsigned short n = c_short(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_int(val buf, val pos, val num)
{
  val self = lit("buf-put-int");
  int n = c_int(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_uint(val buf, val pos, val num)
{
  val self = lit("buf-put-uint");
  unsigned n = c_uint(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_long(val buf, val pos, val num)
{
  val self = lit("buf-put-long");
  long n = c_long(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_ulong(val buf, val pos, val num)
{
  val self = lit("buf-put-ulong");
  unsigned long n = c_ulong(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_float(val buf, val pos, val num)
{
  val self = lit("buf-put-float");
  double n;
  double f = c_flo(num, self);
  if (f > FLT_MAX || f < FLT_MIN)
    uw_throwf(error_s, lit("~a: ~s is out of float range"), self, num, nao);

  n = f;

  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_double(val buf, val pos, val num)
{
  val self = lit("buf-put-double");
  double n = c_flo(num, self);
  buf_put_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num;
}

val buf_put_cptr(val buf, val pos, val cptr)
{
  val self = lit("buf-put-cptr");
  mem_t *p = cptr_get(cptr);
  buf_put_bytes(buf, pos, coerce(mem_t *, &p), sizeof p, self);
  return cptr;
}

void buf_get_bytes(val buf, val pos, mem_t *ptr, ucnum size, val self)
{
  struct buf *b = buf_handle(buf, self);
  ucnum p = buf_check_index(b, pos, self);
  ucnum e = p + size;
  ucnum l = b->len;

  if (e < p)
    err_oflow(self);

  if (e > l)
    uw_throwf(error_s, lit("~a: attempted read past buffer end"), self, nao);

  memcpy(ptr, b->data + p, size);
}

#if HAVE_I8
val buf_get_i8(val buf, val pos)
{
  val self = lit("buf-get-i8");
  struct buf *b = buf_handle(buf, self);
  ucnum p = buf_check_index(b, pos, self);
  if (p >= b->len)
    uw_throwf(error_s, lit("~a: attempted read past buffer end"), self, nao);
  return num_fast(convert(i8_t, b->data[p]));
}

val buf_get_u8(val buf, val pos)
{
  val self = lit("buf-get-u8");
  struct buf *b = buf_handle(buf, self);
  ucnum p = buf_check_index(b, pos, self);
  if (p >= b->len)
    uw_throwf(error_s, lit("~a: attempted read past buffer end"), self, nao);
  return num_fast(convert(u8_t, b->data[p]));
}
#endif

#if HAVE_I16
val buf_get_i16(val buf, val pos)
{
  val self = lit("buf-get-i16");
  i16_t n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num_fast(n);
}

val buf_get_u16(val buf, val pos)
{
  val self = lit("buf-get-u16");
  u16_t n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num_fast(n);
}
#endif

#if HAVE_I32
val buf_get_i32(val buf, val pos)
{
  val self = lit("buf-get-i32");
  i32_t n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num(n);
}

val buf_get_u32(val buf, val pos)
{
  val self = lit("buf-get-u32");
  u32_t n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return unum(n);
}
#endif

#if HAVE_I64
val buf_get_i64(val buf, val pos)
{
  val self = lit("buf-get-i64");
  i64_t n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);

  if (sizeof (i64_t) <= sizeof (cnum)) {
    return num(n);
  } else {
    val high = num(n >> 32);
    val low = unum(n & 0xFFFFFFFF);
    return logior(ash(high, num_fast(32)), low);
  }
}

val buf_get_u64(val buf, val pos)
{
  val self = lit("buf-get-u64");
  u64_t n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);

  if (sizeof (u64_t) <= sizeof (uint_ptr_t)) {
    return unum(n);
  } else {
    val high = unum(n >> 32);
    val low = unum(n & 0xFFFFFFFF);
    return logior(ash(high, num_fast(32)), low);
  }
}
#endif

val buf_get_char(val buf, val pos)
{
#if CHAR_MAX == UCHAR_MAX
  return buf_get_u8(buf, pos);
#else
  return buf_get_i8(buf, pos);
#endif
}

val buf_get_uchar(val buf, val pos)
{
  return buf_get_u8(buf, pos);
}

val buf_get_short(val buf, val pos)
{
  val self = lit("buf-get-short");
  short n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
#if SIZEOF_SHORT < SIZEOF_PTR
  return num_fast(n);
#else
  return num(n);
#endif
}

val buf_get_ushort(val buf, val pos)
{
  val self = lit("buf-get-ushort");
  unsigned short n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
#if SIZEOF_SHORT < SIZEOF_PTR
  return num_fast(n);
#else
  return unum(n);
#endif
}

val buf_get_int(val buf, val pos)
{
  val self = lit("buf-get-int");
  int n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num(n);
}

val buf_get_uint(val buf, val pos)
{
  val self = lit("buf-get-uint");
  unsigned n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return unum(n);
}

val buf_get_long(val buf, val pos)
{
  val self = lit("buf-get-long");
  long n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return num(n);
}

val buf_get_ulong(val buf, val pos)
{
  val self = lit("buf-get-long");
  unsigned long n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return unum(n);
}

val buf_get_float(val buf, val pos)
{
  val self = lit("buf-get-float");
  float n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return flo(n);
}

val buf_get_double(val buf, val pos)
{
  val self = lit("buf-get-double");
  double n;
  buf_get_bytes(buf, pos, coerce(mem_t *, &n), sizeof n, self);
  return flo(n);
}

val buf_get_cptr(val buf, val pos)
{
  val self = lit("buf-get-cptr");
  mem_t *p;
  buf_get_bytes(buf, pos, coerce(mem_t *, &p), sizeof p, self);
  return cptr(p);
}

val buf_print(val buf, val stream_in)
{
  val self = lit("buf-print");
  val stream = default_arg(stream_in, std_output);
  struct buf *b = buf_handle(buf, self);
  ucnum len = b->len, count = 0;
  mem_t *data = b->data;
  val save_mode = test_neq_set_indent_mode(stream, num_fast(indent_foff),
                                           num_fast(indent_data));
  val save_indent;
  int fb = 0;

  put_string(lit("#b'"), stream);

  save_indent = inc_indent(stream, zero);

  while (len-- > 0) {
    format(stream, lit("~,02x"), num_fast(*data++), nao);
    if ((++count & 7) == 0 && len)
      if (width_check(stream, chr(' ')))
        fb = 1;
  }

  set_indent(stream, save_indent);
  set_indent_mode(stream, save_mode);

  if (fb)
    force_break(stream);

  return put_char(chr('\''), stream);
}

val buf_pprint(val buf, val stream_in)
{
  val self = lit("buf-pprint");
  val stream = default_arg(stream_in, std_output);
  struct buf *b = buf_handle(buf, self);
  ucnum len = b->len;
  mem_t *data = b->data;

  if (opt_compat && opt_compat <= 294) {
    while (len-- > 0)
      put_byte(num_fast(*data++), stream);
  } else {
    while (len-- > 0)
      format(stream, lit("~,02x"), num_fast(*data++), nao);
  }

  return t;
}

val buf_str_sep(val buf, val sep, val self)
{
  struct buf *b = buf_handle(buf, self);
  ucnum len = b->len;
  val ret = null_string;

  if (len > 0) {
    val stream = make_string_output_stream();
    ucnum i = 0;

    goto first;

    while (i < len) {
      put_string(sep, stream);
    first:
      format(stream, lit("~,02x"), num_fast(b->data[i++]), nao);
    }

    ret = get_string_from_stream(stream);
  }

  return ret;
}

void buf_hex(val buf, char *hex, size_t sz, int caps)
{
  val self = lit("buf-hex");
  struct buf *b = buf_handle(buf, self);
  size_t i;
  unsigned char *data = b->data;
  const char *fmt = caps ? "%02X" : "%02x";

  *hex = 0;

  for (i = 0; i < sz - 1; i += 2)
    hex += sprintf(hex, fmt, *data++);
}

struct buf_strm {
  struct strm_base a;
  utf8_decoder_t ud;
  int is_byte_oriented;
  val buf;
  ucnum pos;
};

static void buf_strm_mark(val stream)
{
  struct buf_strm *b = coerce(struct buf_strm *, stream->co.handle);
  strm_base_mark(&b->a);
  gc_mark(b->buf);
}

static int buf_strm_put_byte_callback(int b, mem_t *ctx)
{
  struct buf_strm *s = coerce(struct buf_strm *, ctx);
  (void) buf_put_uchar(s->buf, unum(s->pos), num_fast(b));
  s->pos++;
  return 1;
}

static val buf_strm_put_string(val stream, val str)
{
  val self = lit("put-string");
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  const wchar_t *p = c_str(str, self);

  while (*p) {
    (void) utf8_encode(*p++, buf_strm_put_byte_callback, coerce(mem_t *, s));
  }

  return t;
}

static val buf_strm_put_char(val stream, val ch)
{
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  (void) utf8_encode(c_chr(ch), buf_strm_put_byte_callback, coerce(mem_t *, s));
  return t;
}

static val buf_strm_put_byte(val stream, int b)
{
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  (void) buf_strm_put_byte_callback(b, coerce(mem_t *, s));
  return t;
}


static int buf_strm_get_byte_callback(mem_t *ctx)
{
  struct buf_strm *s = coerce(struct buf_strm *, ctx);
  struct buf *b = us_buf_handle(s->buf);
  return (s->pos >= b->len) ? EOF : b->data[s->pos++];
}

static val buf_strm_get_char(val stream)
{
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  wint_t ch;

  if (s->is_byte_oriented) {
    ch = buf_strm_get_byte_callback(coerce(mem_t *, s));
    if (ch == 0)
      ch = 0xDC00;
  } else {
    ch = utf8_decode(&s->ud, buf_strm_get_byte_callback,
                     coerce(mem_t *, s));
  }

  return (ch != WEOF) ? chr(ch) : nil;
}

static val buf_strm_get_byte(val stream)
{
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  int byte = buf_strm_get_byte_callback(coerce(mem_t *, s));
  return byte == EOF ? nil : num_fast(byte);
}

static val buf_strm_unget_char(val stream, val ch)
{
  val self = lit("unget-char");
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  struct buf *b = us_buf_handle(s->buf);
  struct utf8_tiny_buf bu;
  unsigned char *bend = bu.buf + sizeof bu.buf;
  ucnum pos = s->pos;

  bu.ptr = bend;

  (void) utf8_encode(c_chr(ch), utf8_tiny_buf_putc, coerce(mem_t *, &bu));

  if (convert(size_t, bend - bu.ptr) > pos) {
    uw_throwf(file_error_s,
              lit("~a: cannot push back past start of stream ~s"),
              stream, self, nao);
  }

  while (bu.ptr < bend)
    b->data[--pos] = *bu.ptr++;

  s->pos = pos;

  return ch;
}

static val buf_strm_unget_byte(val stream, int byte)
{
  val self = lit("unget-byte");
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  struct buf *b = us_buf_handle(s->buf);
  ucnum pos = s->pos;

  if (pos <= 0) {
    uw_throwf(file_error_s,
              lit("~a: cannot push back past start of stream ~s"),
              self, stream, nao);
  }

  b->data[--pos] = byte;
  s->pos = pos;
  return num_fast(byte);
}

static ucnum buf_strm_fill_buf(val stream, mem_t *ptr, ucnum len, ucnum pos)
{
  val self = lit("fill-buf");
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  struct buf *b = us_buf_handle(s->buf);

  if (convert(size_t, len) != len || len > INT_PTR_MAX)
    uw_throwf(error_s, lit("~a: buffer too large"), self, nao);
  if (pos >= len) {
    return len;
  } else {
    size_t room = len - pos;
    size_t blen = b->size == BUF_BORROWED ? b->len : b->size;
    size_t avail = blen - s->pos;
    size_t copy = min(room, avail);
    memcpy(ptr + pos, b->data + s->pos, copy);
    s->pos += copy;
    return pos + copy;
  }
}


static val buf_strm_seek(val stream, val offset, enum strm_whence whence)
{
  val self = lit("seek-stream");
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  struct buf *b = us_buf_handle(s->buf);
  cnum off = c_num(offset, self);
  ucnum npos;
  ucnum from;

  switch (whence) {
  case strm_start:
    from = 0;
    break;
  case strm_cur:
    if (off == 0)
      return unum(s->pos);
    from = s->pos;
    break;
  case strm_end:
    from = b->len;
    break;
  default:
    internal_error("invalid whence value");
  }

  npos = from + off;

  if ((off >= 0 && npos < from) ||
      (off < 0 && npos > from))
    uw_throwf(file_error_s, lit("~a: overflow in seeking ~s bytes from ~s"),
              self, offset, unum(from), nao);

  if (npos > b->len)
    buf_set_length(s->buf, unum(npos), zero);

  s->pos = npos;
  return t;
}

static val buf_strm_truncate(val stream, val len)
{
  val self = lit("truncate-stream");
  ucnum blen = buf_check_len(len, self);
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  struct buf *b = us_buf_handle(s->buf);

  if (blen >= s->pos) {
    buf_do_set_len(s->buf, b, blen, nil, self);
  } else {
    buf_do_set_len(s->buf, b, s->pos, nil, self);
    memset(b->data + blen, 0, s->pos - blen);
  }

  return t;
}

static val buf_strm_get_prop(val stream, val ind)
{
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);

  if (ind == name_k) {
    return lit("buf-stream");
  } else if (ind == byte_oriented_k) {
    return tnil(s->is_byte_oriented);
  }

  return nil;
}

static val buf_strm_set_prop(val stream, val ind, val prop)
{
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);

  if (ind == byte_oriented_k) {
    s->is_byte_oriented = prop ? 1 : 0;
    return t;
  }

  return nil;
}


static val buf_strm_get_error(val stream)
{
  struct buf_strm *s = coerce(struct buf_strm *, stream->co.handle);
  struct buf *b = us_buf_handle(s->buf);
  return tnil(s->pos >= b->len);
}

static val buf_strm_get_error_str(val stream)
{
  return buf_strm_get_error(stream) ? lit("eof") : lit("no error");
}

static struct strm_ops buf_strm_ops =
  strm_ops_init(cobj_ops_init(eq,
                              stream_print_op,
                              stream_destroy_op,
                              buf_strm_mark,
                              cobj_eq_hash_op,
                              0),
                wli("buf-stream"),
                buf_strm_put_string,
                buf_strm_put_char,
                buf_strm_put_byte,
                generic_get_string,
                buf_strm_get_char,
                buf_strm_get_byte,
                buf_strm_unget_char,
                buf_strm_unget_byte,
                0,
                buf_strm_fill_buf,
                0,
                0,
                buf_strm_seek,
                buf_strm_truncate,
                buf_strm_get_prop,
                buf_strm_set_prop,
                buf_strm_get_error,
                buf_strm_get_error_str,
                0,
                0);

static struct buf_strm *buf_strm(val stream, val self)
{
  struct buf_strm *s = coerce(struct buf_strm *,
                              cobj_handle(self, stream, stream_cls));

  type_assert (stream->co.ops == &buf_strm_ops.cobj_ops,
               (lit("~a: ~a is not a buffer stream"), self, stream, nao));
  return s;
}

val make_buf_stream(val buf_opt)
{
  val self = lit("make-buf-stream");
  val stream;
  val buf = default_arg(buf_opt, make_buf_fast(0, 0, 64));
  struct buf_strm *s = coerce(struct buf_strm *, chk_malloc(sizeof *s));

  (void) buf_handle(buf, self);

  strm_base_init(&s->a);
  utf8_decoder_init(&s->ud);
  s->buf = nil;
  s->pos = 0;
  s->is_byte_oriented = 0;
  stream = cobj(coerce(mem_t *, s), stream_cls, &buf_strm_ops.cobj_ops);
  s->buf = buf;

  return stream;
}

val get_buf_from_stream(val stream)
{
  val self = lit("get-buf-from-stream");
  struct buf_strm *s = buf_strm(stream, self);
  return s->buf;
}

void buf_swap32(val buf)
{
  val self = lit("buf-swap32");
  struct buf *b = buf_handle(buf, self);
  mem_t *data = b->data, *end = data + b->len;

  for (; data + 3 < end; data += 4) {
    u32_t sw32 = *coerce(u32_t *, data);
    sw32 = ((sw32 & 0xFF00FF00U) >>  8) | ((sw32 & 0x00FF00FFU) <<  8);
    sw32 = ((sw32 & 0xFFFF0000U) >> 16) | ((sw32 & 0x0000FFFFU) << 16);
    *coerce(u32_t *, data) = sw32;
  }
}

static val buf_str(val str, val null_term)
{
  val self = lit("buf-str");
  size_t sz;
  val nt = default_null_arg(null_term);
  unsigned char *u8 = utf8_dup_to_buf(c_str(str, self), &sz, nt != nil);
  return make_owned_buf(sz, u8, self);
}

static val str_buf(val buf, val null_term)
{
  val self = lit("str-buf");
  struct buf *b = buf_handle(buf, self);
  val nt = default_null_arg(null_term);
  size_t blen = b->len;
  size_t len = (nt && blen > 0 && !b->data[blen - 1]) ? blen - 1 : blen;
  wchar_t *str = utf8_dup_from_buf(coerce(const char *, b->data), len);
  return string_own(str);
}

static val buf_int(val num)
{
  val self = lit("buf-int");

  switch (type(num)) {
  case CHR:
    return buf_int(num_fast(c_ch(num)));
  case NUM:
    num = bignum(c_num(num, self));
    /* fallthrough */
  case BGNUM:
    {
      val wi = width(num);
      val bits = succ(wi);
      val bytes = ash(plus(bits, num_fast(7)), num_fast(-3));
      val bitsround = ash(bytes, num_fast(3));
      val un = logtrunc(num, bitsround);
      val ube = if3(bignump(un), un, bignum(c_num(un, self)));
      mp_int *m = mp(ube);
      size_t numsize = mp_unsigned_bin_size(m);
      size_t bufsize = c_unum(bytes, self);
      mem_t *data = chk_malloc(bufsize);
      data[0] = 0;
      mp_to_unsigned_bin(m, data + (bufsize - numsize));
      return make_owned_buf(bufsize, data, self);
    }
  default:
    uw_throwf(type_error_s, lit("~a: ~s isn't an integer or character"),
              self, num, nao);
  }
}

static val buf_uint(val num)
{
  val self = lit("buf-uint");

  switch (type(num)) {
  case CHR:
    return buf_uint(num_fast(c_ch(num)));
  case NUM:
    num = bignum(c_num(num, self));
    /* fallthrough */
  case BGNUM:
    {
      mp_int *m = mp(num);
      if (!mp_isneg(m)) {
        size_t size = mp_unsigned_bin_size(m);
        mem_t *data = chk_malloc(size);
        mp_to_unsigned_bin(m, data);
        return make_owned_buf(size, data, self);
      }
    }
    uw_throwf(type_error_s, lit("~a: ~s isn't a non-negative integer"),
              self, num, nao);
  default:
    uw_throwf(type_error_s, lit("~a: ~s isn't an integer or character"),
              self, num, nao);
  }
}

static val int_buf(val buf)
{
  val self = lit("int-buf");
  struct buf *b = buf_handle(buf, self);
  ucnum size = b->len;
  val ubn = make_bignum();
  mp_err mpe = mp_read_unsigned_bin(mp(ubn), b->data, size);
  if (mpe != MP_OKAY)
    do_mp_error(self, mpe);
  return sign_extend(normalize(ubn), unum(b->len * 8));
}

static val uint_buf(val buf)
{
  val self = lit("uint-buf");
  struct buf *b = buf_handle(buf, self);
  ucnum size = b->len;
  val ubn = make_bignum();
  mp_err mpe = mp_read_unsigned_bin(mp(ubn), b->data, size);
  if (mpe != MP_OKAY)
    do_mp_error(self, mpe);
  return normalize(ubn);
}

#if HAVE_ZLIB

static val buf_compress(val buf, val level_opt)
{
  val self = lit("buf-compress");
  val level = default_arg(level_opt, negone);
  int lev = c_int(level, self);
  struct buf *b = buf_handle(buf, self);
  ucnum size = b->len;
  uLong bound = compressBound(size), zsize = bound;
  mem_t *zdata = chk_malloc(bound);

  if (convert(uLong, size) != size) {
    free(zdata);
    err_oflow(self);
  }

  if (compress2(zdata, &zsize, b->data, size, lev) != Z_OK) {
    free(zdata);
    uw_throwf(error_s, lit("~a: compression failed"), self, nao);
  }

  zdata = chk_realloc(zdata, zsize);
  return make_owned_buf(zsize, zdata, self);
}

static val buf_decompress(val buf)
{
  val self = lit("buf-decompress");
  struct buf *b = buf_handle(buf, self);
  ucnum zsize = b->len;
  uLong zsz10 = 10 * zsize;
  uLong size = if3(zsz10 > zsize, zsz10, convert(uLong, -1));
  mem_t *data = chk_malloc(size);

  for (;;) {
    switch (uncompress(data, &size, b->data, zsize)) {
    case Z_OK:
      data = chk_realloc(data, size);
      return make_owned_buf(size, data, self);
    case Z_BUF_ERROR:
      if (size == convert(uLong, -1))
        break;
      if (size * 2 > size)
        size = size * 2;
      else if (size == convert(uLong, -1))
        break;
      else
        size  = convert(uLong, -1);
      data = chk_realloc(data, size);
      continue;
    default:
      break;
    }
    break;
  }

  free(data);
  uw_throwf(error_s, lit("~a: decompression failed"), self, nao);
}

static val str_compress(val str, val level_opt)
{
  val self = lit("str-compress");
  val level = default_arg(level_opt, negone);
  int lev = c_int(level, self);
  size_t sz;
  unsigned char *u8 = utf8_dup_to_buf(c_str(str, self), &sz, 0);
  uLong bound = compressBound(sz), zsize = bound;
  mem_t *zdata = chk_malloc(bound);

  if (convert(uLong, sz) != sz) {
    free(zdata);
    err_oflow(self);
  }

  if (compress2(zdata, &zsize, u8, sz, lev) != Z_OK) {
    free(zdata);
    uw_throwf(error_s, lit("~a: compression failed"), self, nao);
  }

  zdata = chk_realloc(zdata, zsize);
  return make_owned_buf(zsize, zdata, self);
}

static val str_decompress(val buf)
{
  val self = lit("str-decompress");
  struct buf *b = buf_handle(buf, self);
  ucnum zsize = b->len;
  uLong zsz10 = 10 * zsize;
  uLong size = if3(zsz10 > zsize, zsz10, convert(uLong, -1));
  mem_t *data = chk_malloc(size);
  val ret = nil;

  for (;;) {
    switch (uncompress(data, &size, b->data, zsize)) {
    case Z_OK:

      data = chk_realloc(data, size);
      ret = string_utf8_from_buf(coerce(char *, data), size);
      free(data);
      return ret;
    case Z_BUF_ERROR:
      if (size == convert(uLong, -1))
        break;
      if (size * 2 > size)
        size = size * 2;
      else if (size == convert(uLong, -1))
        break;
      else
        size  = convert(uLong, -1);
      data = chk_realloc(data, size);
      continue;
    default:
      break;
    }
    break;
  }

  free(data);
  uw_throwf(error_s, lit("~a: decompression failed"), self, nao);
}

#endif

val buf_ash(val buf, val bits)
{
  val self = lit("buf-ash");
  cnum b = c_num(bits, self);
  struct buf *bh = buf_handle(buf, self);
  ucnum len = bh->len;

  if (b == 0) {
    return buf;
  } else if (b > 0) {
    ucnum r = b % 8;
    ucnum bytes = b / 8;
    ucnum nlen = len + bytes;

    if (nlen < len || nlen > convert(ucnum, INT_PTR_MAX))
      err_oflow(self);

    if (r == 0 || nlen == 0) {
      val nbuf = copy_buf(buf);
      buf_set_length(nbuf, unum(nlen), 0);
      return nbuf;
    } else {
      unsigned acc = 0, b0 = len ? bh->data[0] : 0;
      unsigned c = ((b0 << r) & 0xff00) ? 1 : 0;
      ucnum cnlen = c + nlen;
      val nbuf = make_ubuf(cnlen);
      struct buf *nbh = buf_handle(nbuf, self);
      ucnum i;

      memset(nbh->data + c + len, 0, bytes);

      for (i = len - 1; i != convert(ucnum, -1); --i) {
        unsigned by = bh->data[i] << r;
        nbh->data[i + c] = by | acc;
        acc = by >> 8;
      }

      if (c)
        nbh->data[0] = acc;

      return nbuf;
    }
  } else {
    ucnum r = (-b) % 8;
    ucnum bytes = (-b) / 8;
    ucnum nlen = len >= bytes ? len - bytes : 0;
    unsigned b0 = len ? bh->data[0] : 0;
    unsigned v = b0 && !(b0 >> r);

    if (r == 0 || nlen == 0) {
      if (v && nlen) {
        val nbuf = make_ubuf(nlen - 1);
        struct buf *nbh = buf_handle(nbuf, self);
        memcpy(nbh->data, bh->data + 1, nlen - 1);
        return nbuf;
      } else {
        val nbuf = make_ubuf(nlen);
        struct buf *nbh = buf_handle(nbuf, self);
        memcpy(nbh->data, bh->data, nlen);
        return nbuf;
      }
    } else {
      unsigned acc = v ? b0 << 8 : 0;
      ucnum cnlen = nlen - v;
      val nbuf = make_ubuf(cnlen);
      struct buf *nbh = buf_handle(nbuf, self);
      ucnum i;

      for (i = 0; i < cnlen; i++) {
        unsigned by = bh->data[i + v];
        nbh->data[i] = (by | acc) >> r;
        acc = by << 8;
      }

      return nbuf;
    }
  }
}

val buf_fash(val buf, val bits)
{
  val self = lit("buf-ash");
  cnum b = c_num(bits, self);
  struct buf *bh = buf_handle(buf, self);
  const ucnum len = bh->len;

  if (b == 0 || len == 0) {
    return buf;
  } else if (b > 0) {
    ucnum bytes = b / 8;

    if (bytes >= len) {
      return make_buf_fast(bh->len, 0, bh->len);
    } else {
      ucnum r = b % 8;
      unsigned acc = 0;
      val nbuf = make_ubuf(len);
      struct buf *nbh = buf_handle(nbuf, self);
      ucnum i;

      memset(nbh->data + len - bytes, 0, bytes);

      if (r == 0) {
        memcpy(nbh->data, bh->data + bytes, len - bytes);
      } else for (i = len - bytes - 1; i != convert(ucnum, -1); --i) {
        unsigned by = bh->data[i + bytes] << r;
        nbh->data[i] = by | acc;
        acc = by >> 8;
      }

      return nbuf;
    }
  } else {
    ucnum bytes = (-b) / 8;

    if (bytes >= len) {
      return make_buf_fast(bh->len, 0, bh->len);
    } else {
      ucnum r = (-b) % 8;
      unsigned acc = 0;
      val nbuf = make_ubuf(len);
      struct buf *nbh = buf_handle(nbuf, self);
      ucnum i;

      memset(nbh->data, 0, bytes);

      if (r == 0) {
        memcpy(nbh->data + bytes, bh->data, len - bytes);
      } else for (i = 0; i < len - bytes; i++) {
        unsigned by = bh->data[i];
        nbh->data[i + bytes] = (by | acc) >> r;
        acc = by << 8;
      }

      return nbuf;
    }
  }

  abort();
}

val buf_and(val bufa, val bufb)
{
  val self = lit("buf-and");
  struct buf *ab = buf_handle(bufa, self);
  struct buf *bb = buf_handle(bufb, self);
  ucnum la = ab->len;
  ucnum lb = bb->len;

  if (la == 0) {
    return make_buf_fast(bb->len, 0, bb->len);
  } else if (lb == 0) {
    return make_buf_fast(ab->len, 0, ab->len);
  } else if (la < lb) {
    return buf_and(bufb, bufa);
  } else {
    val obuf = make_ubuf(la);
    ucnum delta = la - lb;
    ucnum i;
    struct buf *ob = buf_handle(obuf, self);

    memset(ob->data, 0, delta);

    for (i = 0; i < lb; i++)
      ob->data[i + delta] = ab->data[i + delta] & bb->data[i];

    return obuf;
  }
}

val buf_test(val bufa, val bufb)
{
  val self = lit("buf-test");
  struct buf *ab = buf_handle(bufa, self);
  struct buf *bb = buf_handle(bufb, self);
  ucnum la = ab->len;
  ucnum lb = bb->len;

  if (la == 0 || lb == 0 ) {
    return nil;
  } else if (la < lb) {
    return buf_test(bufb, bufa);
  } else {
    ucnum delta = la - lb;
    ucnum i;

    for (i = 0; i < lb; i++) {
      if ((ab->data[i + delta] & bb->data[i]) != 0)
        return t;
    }

    return nil;
  }
}

val buf_or(val bufa, val bufb)
{
  val self = lit("buf-or");
  struct buf *ab = buf_handle(bufa, self);
  struct buf *bb = buf_handle(bufb, self);
  ucnum la = ab->len;
  ucnum lb = bb->len;

  if (la == 0) {
    return bufb;
  } else if (lb == 0) {
    return bufa;
  } else if (la < lb) {
    return buf_or(bufb, bufa);
  } else {
    val obuf = make_ubuf(la);
    ucnum delta = la - lb;
    ucnum i;
    struct buf *ob = buf_handle(obuf, self);

    memcpy(ob->data, ab->data, delta);

    for (i = 0; i < lb; i++)
      ob->data[i + delta] = ab->data[i + delta] | bb->data[i];

    return obuf;
  }
}

val buf_xor(val bufa, val bufb)
{
  val self = lit("buf-xor");
  struct buf *ab = buf_handle(bufa, self);
  struct buf *bb = buf_handle(bufb, self);
  ucnum la = ab->len;
  ucnum lb = bb->len;

  if (la == 0) {
    return bufb;
  } else if (lb == 0) {
    return bufa;
  } else if (la < lb) {
    return buf_xor(bufb, bufa);
  } else {
    val obuf = make_ubuf(la);
    ucnum delta = la - lb;
    ucnum i;
    struct buf *ob = buf_handle(obuf, self);

    memcpy(ob->data, ab->data, delta);

    for (i = 0; i < lb; i++)
      ob->data[i + delta] = ab->data[i + delta] ^ bb->data[i];

    return obuf;
  }
}

val buf_not(val buf)
{
  val self = lit("buf-not");
  struct buf *b = buf_handle(buf, self);
  ucnum l = b->len;

  if (l == 0) {
    return buf;
  } else {
    val obuf = make_ubuf(l);
    struct buf *ob = buf_handle(obuf, self);
    ucnum i;

    for (i = 0; i < l; i++)
      ob->data[i] = b->data[i] ^ 0xff;

    return obuf;
  }
}

val buf_trunc(val buf, val bits)
{
  val self = lit("buf-trunc");
  cnum b = c_num(bits, self);
  struct buf *bh = buf_handle(buf, self);
  ucnum l = bh->len;

  if (b <= 0) {
    return make_buf_fast(0, 0, 0);;
  } else {
    ucnum bytes = (b + 7) / 8;
    val obuf = make_ubuf(bytes);
    struct buf *ob = buf_handle(obuf, self);

    if (bytes > l) {
      memcpy(ob->data + bytes - l, bh->data, l);
      memset(ob->data, 0, bytes - l);
    } else {
      memcpy(ob->data, bh->data + l - bytes, bytes);
    }

    if (b % 8 != 0)
      ob->data[0] &= (0xffu >> (8 - b % 8));

    return obuf;
  }
}

val buf_bitset(val buf)
{
  val self = lit("buf-bitset");
  struct buf *b = buf_handle(buf, self);
  ucnum l = b->len, i, j;
  list_collect_decl (out, ptail);

  for (i = l - 1, j = 0; i != convert(ucnum, -1); --i, j++) {
    unsigned byte = b->data[i];
    if (byte) {
      val bitpos = mul(unum(j), num_fast(8));
      if (byte & 1)
        ptail = list_collect(ptail, bitpos);
      if (byte & 2)
        ptail = list_collect(ptail, succ(bitpos));
      if (byte & 4)
        ptail = list_collect(ptail, ssucc(bitpos));
      if (byte & 8)
        ptail = list_collect(ptail, sssucc(bitpos));
      if (byte & 16)
        ptail = list_collect(ptail, plus(bitpos, num_fast(4)));
      if (byte & 32)
        ptail = list_collect(ptail, plus(bitpos, num_fast(5)));
      if (byte & 64)
        ptail = list_collect(ptail, plus(bitpos, num_fast(6)));
      if (byte & 128)
        ptail = list_collect(ptail, plus(bitpos, num_fast(7)));
    }
  }

  return out;
}

val buf_bit(val buf, val bit)
{
  val self = lit("buf-bit");
  struct buf *bh = buf_handle(buf, self);
  ucnum l = bh->len;
  cnum b = c_num(bit, self);

  if (b < 0) {
    return nil;
  } else {
    ucnum byte = b / 8;
    ucnum r = b % 8;

    if (byte >= l)
      return nil;

    return tnil(bh->data[l - byte - 1] & (1U << r));
  }
}

val buf_zero_p(val buf)
{
  val self = lit("buf-zero-p");
  struct buf *b = buf_handle(buf, self);
  ucnum l = b->len, i;
  ucnum *ucdata = coerce(ucnum *, b->data);

  for (i = 0; i < l / sizeof (ucnum); i++) {
    if (ucdata[i])
      return nil;
  }

  for (i = l / sizeof (ucnum) * sizeof (ucnum); i < l; i++)
    if (b->data[i])
      return nil;

  return t;
}

val buf_count_ones(val buf)
{
  val self = lit("buf-count-ones");
  struct buf *b = buf_handle(buf, self);
  ucnum l = b->len, i;
  ucnum *ucdata = coerce(ucnum *, b->data);
  ucnum total[2] = { 0, 0 };

  for (i = 0; i < l / sizeof (ucnum); i++) {
    ucnum d = ucdata[i];

    if (d == 0) {
      continue;
    } else if (d == convert(ucnum, -1)) {
      d = sizeof(ucnum) * 8;
    } else if ((d & (d - 1)) == 0) {
      d = 1;
    } else {
#if SIZEOF_PTR == 8
      d = ((d & 0xAAAAAAAAAAAAAAAA) >>  1) + (d & 0x5555555555555555);
      d = ((d & 0xCCCCCCCCCCCCCCCC) >>  2) + (d & 0x3333333333333333);
      d = ((d & 0xF0F0F0F0F0F0F0F0) >>  4) + (d & 0x0F0F0F0F0F0F0F0F);
      d = ((d & 0xFF00FF00FF00FF00) >>  8) + (d & 0x00FF00FF00FF00FF);
      d = ((d & 0xFFFF0000FFFF0000) >> 16) + (d & 0x0000FFFF0000FFFF);
      d = ((d & 0xFFFFFFFF00000000) >> 32) + (d & 0x00000000FFFFFFFF);
#elif SIZEOF_PTR == 4
      d = ((d & 0xAAAAAAAA) >>  1) + (d & 0x55555555);
      d = ((d & 0xCCCCCCCC) >>  2) + (d & 0x33333333);
      d = ((d & 0xF0F0F0F0) >>  4) + (d & 0x0F0F0F0F);
      d = ((d & 0xFF00FF00) >>  8) + (d & 0x00FF00FF);
      d = ((d & 0xFFFF0000) >> 16) + (d & 0x0000FFFF);
#else
#error portme
#endif
    }
    total[1] += d;
    if (total[1] < d)
      total[0]++;
  }

  for (i = l / sizeof (ucnum) * sizeof (ucnum); i < l; i++) {
    unsigned d = b->data[i];

    if (!d) {
      continue;
    } else if (d == 255) {
      d = 8;
    } else if ((d & (d - 1)) == 0) {
      d = 1;
    } else {
      d = ((d & 0xAA) >> 1) + (d & 0x55);
      d = ((d & 0xCC) >> 2) + (d & 0x33);
      d = ((d & 0xF0) >> 4) + (d & 0x0F);
    }

    total[1] += d;
    if (total[1] < d)
      total[0]++;
  }

  return total[0]
         ? plus(ash(unum(total[0]), num_fast(8 * SIZEOF_PTR)), unum(total[1]))
         : unum(total[1]);
}

val buf_binary_width(val buf)
{
  val self = lit("buf-bit-width");
  struct buf *b = buf_handle(buf, self);
  ucnum l = b->len, i;
  ucnum *ucdata = coerce(ucnum *, b->data);
  ucnum zeros[2] = { 0, 0 };
  int found = 0;

  for (i = 0; !found && i < l / sizeof (ucnum); i++) {
    ucnum d = ucdata[i];
    unsigned z = sizeof (ucnum) * 8;

#if !HAVE_LITTLE_ENDIAN
    if (d != 0) {
      found = 1;
      if (d <= INT_PTR_MAX) {
        z = sizeof (ucnum) * 8 - highest_bit(d);
      } else {
        z = 0;
      }
    }
#else
    if (d != 0) {
      found = 1;
      if ((d & 0xff) <= 0x7f) {
#if SIZEOF_PTR == 8
        d = (d << 32) | (d >> 32);
        d = ((d & 0x0000FFFF0000FFFF) << 16) | ((d & 0xFFFF0000FFFF0000) >> 16);
        d = ((d & 0x00FF00FF00FF00FF) << 8) | ((d & 0xFF00FF00FF00FF00) >> 8);
#elif SIZEOF_PTR == 4
        d = (d << 16) | (d >> 16);
        d = ((d & 0x00FF00FF) << 8) | ((d & 0xFF00FF00) >> 8);
#else
#error portme
#endif
        z = sizeof (ucnum) * 8 - highest_bit(d);
      } else {
        z = 0;
      }
    }
#endif

    zeros[1] += z;
    if (zeros[1] < z)
      zeros[0]++;
  }

  for (i = l / sizeof (ucnum) * sizeof (ucnum); !found && i < l; i++) {
    unsigned d = b->data[i];
    unsigned z = 8;

    if (d != 0) {
      found = 1;

      if ((d & 0xFE) == 0)
        z = 7;
      else if ((d & 0xFC) == 0)
        z = 6;
      else if ((d & 0xF8) == 0)
        z = 5;
      else if ((d & 0xF0) == 0)
        z = 4;
      else if ((d & 0xE0) == 0)
        z = 3;
      else if ((d & 0xC0) == 0)
        z = 2;
      else if ((d & 0x80) == 0)
        z = 1;
      else
        z = 0;
    }

    zeros[1] += z;
    if (zeros[1] < z)
      zeros[0]++;
  }

  {
    val tbits = mul(unum(l), num_fast(8));
    val tzeros = zeros[0]
                ? plus(ash(unum(zeros[0]), num_fast(8 * SIZEOF_PTR)),
                       unum(zeros[1]))
                : unum(zeros[1]);
    return minus(tbits, tzeros);
  }
}

val buf_xor_pattern(val buf, val pat)
{
  val self = lit("buf-xor-pattern");
  struct buf *b = buf_handle(buf, self);
  ucnum l = b->len, pl;
  mem_t *p = 0, *fr = 0;
  unsigned char pbuf[SIZEOF_PTR] = { 0 };
  int left = 0;

  prepare_pattern(&p, &fr, pbuf, &pl, &left, pat, self);

  {
    val nbuf = make_ubuf(l);
    struct buf *nbh = buf_handle(nbuf, self);
    ucnum i, j;

    if (left) {
      for (i = 0, j = 0; i < l; i++) {
        nbh->data[i] = b->data[i] ^ p[j++];
        if (j >= pl)
          j = 0;
      }
    } else {
      for (i = l - 1, j = pl; i != convert(ucnum, -1); --i) {
        nbh->data[i] = b->data[i] ^ p[--j];
        if (j == 0)
          j = pl;
      }
    }

    free(fr);
    gc_hint(pat);
    return nbuf;
  }
}

void buf_init(void)
{
  reg_fun(intern(lit("make-buf"), user_package), func_n3o(make_buf, 1));
  reg_fun(intern(lit("bufp"), user_package), func_n1(bufp));
  reg_fun(intern(lit("buf-trim"), user_package), func_n1(buf_trim));
  reg_fun(intern(lit("buf-set-length"), user_package), func_n3o(buf_set_length, 2));
  reg_fun(intern(lit("length-buf"), user_package), func_n1(length_buf));
  reg_fun(intern(lit("buf-alloc-size"), user_package), func_n1(buf_alloc_size));
  reg_fun(intern(lit("copy-buf"), user_package), func_n1(copy_buf));
  reg_fun(intern(lit("sub-buf"), user_package), func_n3(sub_buf));
  reg_fun(intern(lit("replace-buf"), user_package), func_n4(replace_buf));
  reg_fun(intern(lit("buf-list"), user_package), func_n1(buf_list));
  reg_fun(intern(lit("buf-put-buf"), user_package), func_n3(compat_buf_put_buf));

#if HAVE_I8
  reg_fun(intern(lit("buf-put-i8"), user_package), func_n3(buf_put_i8));
  reg_fun(intern(lit("buf-put-u8"), user_package), func_n3(buf_put_u8));
#endif

#if HAVE_I16
  reg_fun(intern(lit("buf-put-i16"), user_package), func_n3(buf_put_i16));
  reg_fun(intern(lit("buf-put-u16"), user_package), func_n3(buf_put_u16));
#endif

#if HAVE_I32
  reg_fun(intern(lit("buf-put-i32"), user_package), func_n3(buf_put_i32));
  reg_fun(intern(lit("buf-put-u32"), user_package), func_n3(buf_put_u32));
#endif

#if HAVE_I64
  reg_fun(intern(lit("buf-put-i64"), user_package), func_n3(buf_put_i64));
  reg_fun(intern(lit("buf-put-u64"), user_package), func_n3(buf_put_u64));
#endif

  reg_fun(intern(lit("buf-put-char"), user_package), func_n3(buf_put_char));
  reg_fun(intern(lit("buf-put-uchar"), user_package), func_n3(buf_put_uchar));
  reg_fun(intern(lit("buf-put-short"), user_package), func_n3(buf_put_short));
  reg_fun(intern(lit("buf-put-ushort"), user_package), func_n3(buf_put_ushort));
  reg_fun(intern(lit("buf-put-int"), user_package), func_n3(buf_put_int));
  reg_fun(intern(lit("buf-put-uint"), user_package), func_n3(buf_put_uint));
  reg_fun(intern(lit("buf-put-long"), user_package), func_n3(buf_put_long));
  reg_fun(intern(lit("buf-put-ulong"), user_package), func_n3(buf_put_ulong));
  reg_fun(intern(lit("buf-put-float"), user_package), func_n3(buf_put_float));
  reg_fun(intern(lit("buf-put-double"), user_package), func_n3(buf_put_double));
  reg_fun(intern(lit("buf-put-cptr"), user_package), func_n3(buf_put_cptr));

#if HAVE_I8
  reg_fun(intern(lit("buf-get-i8"), user_package), func_n2(buf_get_i8));
  reg_fun(intern(lit("buf-get-u8"), user_package), func_n2(buf_get_u8));
#endif

#if HAVE_I16
  reg_fun(intern(lit("buf-get-i16"), user_package), func_n2(buf_get_i16));
  reg_fun(intern(lit("buf-get-u16"), user_package), func_n2(buf_get_u16));
#endif

#if HAVE_I32
  reg_fun(intern(lit("buf-get-i32"), user_package), func_n2(buf_get_i32));
  reg_fun(intern(lit("buf-get-u32"), user_package), func_n2(buf_get_u32));
#endif

#if HAVE_I64
  reg_fun(intern(lit("buf-get-i64"), user_package), func_n2(buf_get_i64));
  reg_fun(intern(lit("buf-get-u64"), user_package), func_n2(buf_get_u64));
#endif

  reg_fun(intern(lit("buf-get-char"), user_package), func_n2(buf_get_char));
  reg_fun(intern(lit("buf-get-uchar"), user_package), func_n2(buf_get_uchar));
  reg_fun(intern(lit("buf-get-short"), user_package), func_n2(buf_get_short));
  reg_fun(intern(lit("buf-get-ushort"), user_package), func_n2(buf_get_ushort));
  reg_fun(intern(lit("buf-get-int"), user_package), func_n2(buf_get_int));
  reg_fun(intern(lit("buf-get-uint"), user_package), func_n2(buf_get_uint));
  reg_fun(intern(lit("buf-get-long"), user_package), func_n2(buf_get_long));
  reg_fun(intern(lit("buf-get-ulong"), user_package), func_n2(buf_get_ulong));
  reg_fun(intern(lit("buf-get-float"), user_package), func_n2(buf_get_float));
  reg_fun(intern(lit("buf-get-double"), user_package), func_n2(buf_get_double));
  reg_fun(intern(lit("buf-get-cptr"), user_package), func_n2(buf_get_cptr));

  reg_fun(intern(lit("make-buf-stream"), user_package), func_n1o(make_buf_stream, 0));
  reg_fun(intern(lit("get-buf-from-stream"), user_package), func_n1(get_buf_from_stream));

  reg_fun(intern(lit("buf-str"), user_package), func_n2o(buf_str, 1));
  reg_fun(intern(lit("str-buf"), user_package), func_n2o(str_buf, 1));
  reg_fun(intern(lit("buf-int"), user_package), func_n1(buf_int));
  reg_fun(intern(lit("buf-uint"), user_package), func_n1(buf_uint));
  reg_fun(intern(lit("int-buf"), user_package), func_n1(int_buf));
  reg_fun(intern(lit("uint-buf"), user_package), func_n1(uint_buf));

#if HAVE_ZLIB
  reg_fun(intern(lit("buf-compress"), user_package), func_n2o(buf_compress, 1));
  reg_fun(intern(lit("buf-decompress"), user_package), func_n1(buf_decompress));
  reg_fun(intern(lit("str-compress"), user_package), func_n2o(str_compress, 1));
  reg_fun(intern(lit("str-decompress"), user_package), func_n1(str_decompress));
#endif

  reg_fun(intern(lit("buf-ash"), user_package), func_n2(buf_ash));
  reg_fun(intern(lit("buf-fash"), user_package), func_n2(buf_fash));
  reg_fun(intern(lit("buf-and"), user_package), func_n2(buf_and));
  reg_fun(intern(lit("buf-test"), user_package), func_n2(buf_test));
  reg_fun(intern(lit("buf-or"), user_package), func_n2(buf_or));
  reg_fun(intern(lit("buf-xor"), user_package), func_n2(buf_xor));
  reg_fun(intern(lit("buf-not"), user_package), func_n1(buf_not));
  reg_fun(intern(lit("buf-trunc"), user_package), func_n2(buf_trunc));
  reg_fun(intern(lit("buf-bitset"), user_package), func_n1(buf_bitset));
  reg_fun(intern(lit("buf-bit"), user_package), func_n2(buf_bit));
  reg_fun(intern(lit("buf-zero-p"), user_package), func_n1(buf_zero_p));
  reg_fun(intern(lit("buf-count-ones"), user_package), func_n1(buf_count_ones));
  reg_fun(intern(lit("buf-binary-width"), user_package), func_n1(buf_binary_width));
  reg_fun(intern(lit("buf-xor-pattern"), user_package), func_n2(buf_xor_pattern));

  fill_stream_ops(&buf_strm_ops);
}
